/* Mednafen Network Play Server
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 TODO to protocol 4:
	Revamped login process:
		Give protocol-version-agnostic error message on unsupported protocol versions.

		<other stuff>

 Next version TODO:
	More verbose quit messages(sent to remaining players).

	Better exception handling - different exceptions for client-specific error and game-end "error".
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>

#ifdef HAVE_MLOCKALL
#include <sys/mman.h>
#endif

#include <assert.h>

#include <exception>

#include "types.h"
#include "md5.h"
#include "time64.h"
#include "trim.inc"

#ifndef SOL_TCP
#define 	SOL_TCP   IPPROTO_TCP
#endif

static const int MaxNickLen = 150; 		// In bytes, not glyphs(worst-case max glyphs using UTF8 is probably around MaxNickLen / 5).
static const int MaxClientsPerGame = 32;	// Maximum number of clients per game.  DO NOT CHANGE TO A HIGHER NUMBER WITHOUT FINDING
						// AND FIXING BITFIELDS.
static const int MaxControllersPerGame = 16;	// Maximum number of controllers per game.  Values higher than 16 are currently illegal.

static const int MaxTotalControllersDataSize = 	512; // Maximum data size for all the controller data per game.
						// Assume usable-worst-case 12 bytes per controller(in case of mouse, x delta, y delta, buttons), a maximum of 8
						// controllers, would give us 96 bytes.
						// We'll be extra-generous and set it to 512 bytes(in case anyone is goofy enough to try to use
						// the Family Keyboard over netplay or something).

// TODO: Ignoring incoming save state receives for a game if a previous save state transfer is in progress.  Or perhaps, if any sendq is >= MaxCommandPayload
// for the game, ignore all large incoming data transfers.

// TODO: Save state transfer throttling.

// TODO: Text throttling.

// TODO: Connection throttling.

// TODO: Nick change throttling.


// 0x40 ... 0x7F are used for payload-less netplay-specific commands(or when the payload can fit in the 4-byte length field).
#define MDFNNPCMD_SETFPS        0x40 /* Sent from client to server ONLY(it should be ignored server-side if it's not from the first
                                        active player for the game). */
#define MDFNNPCMD_NOP		0x41
//
//
//
#define MDFNNPCMD_TEXT          0x90
#define MDFNNPCMD_SERVERTEXT    0x93 // Server text message(informational), server->client (TODO;;; Restrict to protocol 3 and higher).
#define MDFNNPCMD_ECHO          0x94 // Echos the string(no larger than 256 bytes) back to the client(used for pinging).
#define MDFNNPCMD_INTEGRITY     0x95 // Send from a client to a server, then from the server to all clients.
#define MDFNNPCMD_INTEGRITY_RES 0x96 // Integrity result, sent from the clients to the server.  The result should be no larger
                                     // than 256 bytes.
#define MDFNNPCMD_SETNICK       0x98 /* Sent from client to server only. */
#define MDFNNPCMD_PLAYERJOINED  0xA0    // Data:  <byte: bitmask, which inputs this player controls>
                                        //        <bytestream: nickname>
#define MDFNNPCMD_PLAYERLEFT    0xA1    // Data: (see above)
#define MDFNNPCMD_YOUJOINED     0xB0
#define MDFNNPCMD_YOULEFT       0xB1

#define MDFNNPCMD_NICKCHANGED   0xB8    // Sent from server to client

struct GameEntry;

struct ClientEntry
{
	uint32 id;	/* mainly for faster referencing when pointed to from the Games
		           entries.
			 */
	char nickname[MaxNickLen + 1];

	int TCPSocket;
	GameEntry *game;	/* Pointer to the game this player is in. */
	int local_players;	/* The number of local players, 1 through total_controllers */

	time_t timeconnect;	/* Time the client made the connection. */


	uint8 pending_command;	/* Only for protocol >= 1 */
	int total_controllers;  /* Total controllers for the emulated system. */
	int controller_data_size[MaxControllersPerGame]; /* Controller data size. */
	int total_controllers_data_size; /* Data size of all controllers */
	int local_controllers_data_size; /* Data size of local controllers */
	int controller_data_offset[MaxControllersPerGame];
	int merge_target;	/* 0 if not active, 1 = player 1, 2 = player 2, etc. */

	int protocol_version;	/* 0 = FCE Ultra + Mednafen < 0.5.0 - NOT SUPPORTED*/
				/* 1 = Mednafen >= 0.5.0 
				    Changes:  
					1 extra byte in incoming data packet to denote command, instead
					of using 0xFF in the first joystick byte as an escape.
				   2 = Mednafen >= 0.7.0
				    Changes:
					Support for a wider variety of input devices.
				*/

	/* Variables to handle non-blocking TCP reads. */
	uint8 *nbtcp;
	uint32 nbtcphas, nbtcplen;
	uint32 nbtcptype;
	int64 last_receive_time;	// In microseconds

	uint8 *sendq;
	uint32 sendqlen;
	uint32 sendqalloced;
};

struct GameEntry
{
	/* Unique 128-bit identifier for this game session. */
        uint8 id[16];

	/* The version of the protocol the clients in this game are communicating with. */
	int ProtocolVersion;

	/* Text string(not necessarily null-terminated) representing the emulator being used, and the version. */
	uint8 EmulatorID[64];

	/* Total controllers for this game. */
	int TotalControllers;

	/* Data size for each controller. */
	int ControllerDataSize[MaxControllersPerGame];

	/* Total data size for all controllers. */
        int TotalControllersDataSize;

	/* Offset into joybuf for each controller. */
	int ControllerDataOffset[MaxControllersPerGame];

	/* Pointers to connected clients. */
        ClientEntry *Clients[MaxClientsPerGame];

	int IsUnique[MaxClientsPerGame];
		 		/* Set to 1 if player is unique client, 0
				   if it's a duplicated entry to handle multiple players
				   per client.
				*/
	int64 last_time;
	uint32 fps;

        uint8 joybuf[MaxTotalControllersDataSize + 1]; /* X player data + 1 command byte8 */
};

struct CONFIG
{
	int MaxClients;		/* The maximum number of clients to allow. */
				/* You should expect each client to use
				   65-70Kbps(not including save state loads) while
				   connected(and logged in).
				*/

	int ConnectTimeout;	/* How long to wait(in seconds) for the client to provide
				   login data before disconnecting the client.
				*/

	int IdleTimeout;	/* If the data is not received from the client in this amount of time(in seconds), disconnect the client.
				*/

	int Port;		/* The port to listen on. */
	uint8 *Password;	/* The server password. */

	uint32 MinSendQSize;	//
	uint32 MaxSendQSize;	// Maximum size the internal soft send queue is allowed to grow to.  If it grows
                                // larger than this, the connection will be dropped.  Per client limit.
	uint32 MaxCommandPayload;
};

CONFIG ServerConfig;

int LoadConfig(char *fn)
{
 FILE *fp;

 ServerConfig.Port = -1;
 ServerConfig.MaxClients = -1;
 ServerConfig.ConnectTimeout = -1;

 // Settings with defaults(can be missing from config file):
 ServerConfig.IdleTimeout = 30;
 ServerConfig.MinSendQSize = 262144;
 ServerConfig.MaxSendQSize = 8388608;
 ServerConfig.MaxCommandPayload = 5 * 1024 * 1024;

 if((fp=fopen(fn,"rb")))
 {
  char buf[512];
  char sname[sizeof(buf)];
  char args[sizeof(buf)];

  while(fgets(buf, sizeof(buf), fp) > 0)
  {
   char *sc_pos;

   // Handle comments
   if((sc_pos = strchr(buf, ';')))
    *sc_pos = 0;

   trim(buf);

   if(buf[0] == 0)
    continue;

   sscanf(buf, "%s %[^\n]", sname, args);
   trim(args);

   if(!strncasecmp(sname, "maxclients",strlen("maxclients")))
    sscanf(args, "%d",&ServerConfig.MaxClients);
   else if(!strncasecmp(sname, "connecttimeout ",strlen("connecttimeout")))
    sscanf(args, "%d",&ServerConfig.ConnectTimeout);
   else if(!strncasecmp(sname, "port",strlen("port")))
   {
    sscanf(args, "%d",&ServerConfig.Port);
   }
   else if(!strncasecmp(sname, "password",strlen("password")))
   {
    if(args[0] != 0)
    {
     struct md5_context md5;
     ServerConfig.Password = (uint8 *)malloc(16);
     md5_starts(&md5);
     md5_update(&md5,(uint8*)args,strlen(args));
     md5_finish(&md5,ServerConfig.Password);
     puts("Password required to log in.");
    }
   }
   else if(!strncasecmp(sname, "idletimeout", strlen("idletimeout")))
    sscanf(args, "%u", &ServerConfig.IdleTimeout);
   else if(!strncasecmp(sname, "minsendqsize", strlen("minsendqsize")))
    sscanf(args, "%u", &ServerConfig.MinSendQSize);
   else if(!strncasecmp(sname, "maxsendqsize", strlen("maxsendqsize")))
    sscanf(args, "%u", &ServerConfig.MaxSendQSize);
   else if(!strncasecmp(sname, "maxcmdpayload", strlen("maxcmdpayload")))
    sscanf(args, "%u", &ServerConfig.MaxCommandPayload);
   else
   {
    printf("Unknown directive in configuration file: %s\n", sname);
    return(0);
   }
  }
 }
 else return(0);

 if(ServerConfig.Port == -1 || ServerConfig.MaxClients == -1 || ServerConfig.ConnectTimeout == -1)
 {
  puts("Incomplete configuration file");
  return(0);
 }


 printf("Server configuration:\n");
 printf("\tMaximum clients: %u\n", ServerConfig.MaxClients);
 printf("\tConnect timeout: %u seconds\n", ServerConfig.ConnectTimeout);
 printf("\tListen port: %u\n", ServerConfig.Port);
 printf("\tPassword: %s\n", ServerConfig.Password ? "(used)" : "(unused)");
 printf("\tIdle timeout: %u seconds\n", ServerConfig.IdleTimeout);
 printf("\tMinimum internal send queue size: %u bytes\n", ServerConfig.MinSendQSize);
 printf("\tMaximum internal send queue size: %u bytes\n", ServerConfig.MaxSendQSize);
 printf("\tMaximum command payload size: %u bytes\n", ServerConfig.MaxCommandPayload);
 printf("\t---------------------------------------\n");
 printf("\tRough worst-case internal queue memory usage: %.2f MiB\n", ((double)ServerConfig.MaxSendQSize + ServerConfig.MaxCommandPayload) * ServerConfig.MaxClients / 1024 / 1024);
 printf("\n");

 return(1);
}

static ClientEntry *AllClients;
static GameEntry *Games;

static void en32(uint8 *buf, uint32 morp)
{
 buf[0]=morp;
 buf[1]=morp>>8;
 buf[2]=morp>>16;
 buf[3]=morp>>24;
}

static uint32 de32(const uint8 *morp)
{
 return(morp[0]|(morp[1]<<8)|(morp[2]<<16)|(morp[3]<<24));
}

static int MakeSendTCP(ClientEntry *client, const uint8 *data, uint32 len) throw(int);
static void CleanNick(char *nick);
static bool NickUnique(ClientEntry *client, const char *);
static void ParseAndSetNickname(ClientEntry *client, const uint8 *data, uint32 data_len, bool is_change = false);

static void AddClientToGame(ClientEntry *client, const uint8 id[16], const uint8 *emu_id) throw(int);
static void SendCommand(ClientEntry *client, uint8 cmd, const uint8 *data, uint32 len) throw(int);
static void SendToAllInSameGame(ClientEntry *game, int cmd, const uint8 *data, uint32 len) throw(int);
static void TextToClient(ClientEntry *client, const char *fmt, ...) throw(int);
static void AnnouncePlayer(ClientEntry *client, ClientEntry *newclient) throw(int);
static void AnnouncePlayerLeft(ClientEntry *client, ClientEntry *newclient, uint32 mps, uint32 mps_merged) throw(int);
static void KillClient(ClientEntry *client);

#define NBTCP_LOGINLEN		0x100
#define NBTCP_LOGIN		0x200
#define NBTCP_COMMANDLEN	0x300
#define NBTCP_COMMAND		0x400

#define NBTCP_UPDATEDATA	0x800

static void StartNBTCPReceive(ClientEntry *client, uint32 type, uint32 len)
{
 client->nbtcp = (uint8 *)malloc(len);
 client->nbtcplen = len;
 client->nbtcphas = 0;
 client->nbtcptype = type;
}

static void RedoNBTCPReceive(ClientEntry *client)
{
 client->nbtcphas = 0;
}

static void EndNBTCPReceive(ClientEntry *client)
{
    free(client->nbtcp);
    client->nbtcplen = 0;
    client->nbtcphas = 0;
    client->nbtcptype = 0;
    client->nbtcp = 0;
}

static uint32 MakeMPS(ClientEntry *client, uint32 *mergedto)
{
 uint32 ret = 0;
 GameEntry *game = client->game;

 if(mergedto != NULL)
 {
  *mergedto = 0;

  if(client->merge_target)
   *mergedto |= 1 << (client->merge_target - 1);
 }

 for(int x = 0; x < game->TotalControllers; x++)
 {
  if(game->Clients[x] == client)
  {
   ret |= 1U << x;
  }
 }
 return(ret);
}

// Uses static buffer, so copy/display the data immediately.
static char *MakeMPSString(ClientEntry *client)
{
 static char buf[MaxNickLen + 1];
 GameEntry *game = client->game;
 uint32 mps;
 uint32 mergedto;

 mps = MakeMPS(client, &mergedto);

 if(!mps)
 {
  buf[0] = 0;

  // FIXME, this is silly.
  for(int x = 0; x < MaxClientsPerGame; x++)
  {
   if(game->Clients[x] == client)
   {
    snprintf(buf, sizeof(buf), "*Lurker %d", x + 1 - game->TotalControllers);
    break;
   }
  }
 }
 else
 {
  snprintf(buf, sizeof(buf), "*Player %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                                       (mps & 0x0001) ? " 1" : "",
                                       (mps & 0x0002) ? " 2" : "",
                                       (mps & 0x0004) ? " 3" : "",
                                       (mps & 0x0008) ? " 4" : "",
                                       (mps & 0x0010) ? " 5" : "",
                                       (mps & 0x0020) ? " 6" : "",
                                       (mps & 0x0040) ? " 7" : "",
                                       (mps & 0x0080) ? " 8" : "",
                                       (mps & 0x0100) ? " 9" : "",
                                       (mps & 0x0200) ? " 10" : "",
                                       (mps & 0x0400) ? " 11" : "",
                                       (mps & 0x0800) ? " 12" : "",
                                       (mps & 0x1000) ? " 13" : "",
                                       (mps & 0x2000) ? " 14" : "",
                                       (mps & 0x4000) ? " 15" : "",
                                       (mps & 0x8000) ? " 16" : "");
 }

 if(client->merge_target)
 {
  snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " MT: %d", client->merge_target);
 }

 return(buf);
}

struct login_data_t
{
 uint8 gameid[16];
 uint8 password[16];

 union
 {
  struct
  {
   uint8 protocol_version;
   uint8 total_controllers;
   uint8 padding0[2];

   uint8 pd_name_len[4];		// (protocol 2.  port device name len.  Deprecated, never implemented server-side) LSB first, 32-bit
					// Protocol 3, length of emulator name and version string(up to 64 bytes) - (note that any non-0
					// bytes < 0x20 in this string will be replaced by 0x20).

   uint8 padding1[8];

   uint8 controller_data_size[16];	// Only valid for protocol >= 2.

   uint8 merge_target;

   uint8 padding2[15];

   uint8 controller_type[16];		// Protocol >= ?(TODO)
  };

  uint8 extra[64];
 };

 uint8 local_players;
 uint8 nickname[0];
};


#if 0
// Ideas:
//  use MSB of login len to denote Protocol 4 and higher
//
struct p4_login_data_t
{
 uint8 magic[16]; 		// string 	- "MEDNAFEN_NETPLAY"
 uint8 protocol_version[4];	// uint32 	- Protocol version
 uint8 password[16];		//
 uint8 nickname[256];

 // Game configuration data.
 uint8 game_id[32];		//
 uint8 emu_id[64];		//		- Unique emulator identifier incorporating name and version.
 uint8 emu_config[2048];	//		- Free-form, emulator-specific emulator configuration data.
 uint8 total_controllers;
 uint8 controller_data_size[32];

 // Player-specific game configuration data.
 uint8 local_players;
 uint8 merge_target;
};
#endif

/* Returns 1 if we are back to normal game mode, 0 if more data is yet to arrive. */
static int CheckNBTCPReceive(ClientEntry *client) throw(int)
{
 if(!client->nbtcplen)
  throw(1);			/* Should not happen. */
 int l;
       
 while((l = recv(client->TCPSocket, client->nbtcp + client->nbtcphas, client->nbtcplen  - client->nbtcphas, MSG_NOSIGNAL)))
 {
  if(l == -1)
  {
   if(errno == EAGAIN || errno == EWOULDBLOCK)
    break;
   throw(1); /* Die now.  NOW. */
  }

  assert(l != 0);
  client->last_receive_time = MBL_Time64();

  client->nbtcphas += l;

  //printf("Read: %d, %04x, %d, %d\n",l,client->nbtcptype,client->nbtcphas, client->nbtcplen);

  /* We're all full.  Yippie. */
  if(client->nbtcphas == client->nbtcplen)
  {
   switch(client->nbtcptype & 0xF00)
   {
    case NBTCP_UPDATEDATA:
			{
			 GameEntry *game = client->game;
		         int x, wx;

			 if(client->nbtcp[0])
			 {
			  client->pending_command = client->nbtcp[0];
			  EndNBTCPReceive(client);
			  StartNBTCPReceive(client, NBTCP_COMMANDLEN, 4);
			  return(1);
			 }

		         for(x=0,wx=0; x < game->TotalControllers; x++)
		         {
		          if(game->Clients[x] == client)
		          {
			   memcpy(&game->joybuf[game->ControllerDataOffset[x]], &client->nbtcp[1 + wx], game->ControllerDataSize[x]);
		           wx += game->ControllerDataSize[x];
		          }
			 }
			 RedoNBTCPReceive(client);
			}
			return(1);

    case NBTCP_COMMANDLEN:
			{
			 uint8 cmd = client->pending_command;
			 uint32 len = de32(client->nbtcp);

			 if((cmd & 0x80) && len > ServerConfig.MaxCommandPayload)	/* Sanity check. */
			 {
		          printf("Client %d on game %d wants to exceed maximum command payload by %u bytes!\n", client->id, (int)(client->game-Games), len - ServerConfig.MaxCommandPayload);
			  throw(1);
			 }

			 if(!(cmd & 0x80))	// No payload, or payload is length data.
		         {
			  if(cmd == MDFNNPCMD_SETFPS)
                          {
                           client->game->fps = len;
                           EndNBTCPReceive(client);
                           StartNBTCPReceive(client, NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
                          }
			  else if(cmd == MDFNNPCMD_NOP)
			  {
                           EndNBTCPReceive(client);
                           StartNBTCPReceive(client, NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			  }
		          else
			  {
			   SendToAllInSameGame(client, cmd, 0, 0);
			   EndNBTCPReceive(client);
			   StartNBTCPReceive(client, NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			  }
			 }
			 else // Everything else.
			 {
			  if(cmd == MDFNNPCMD_INTEGRITY)
                          {
                           SendToAllInSameGame(client, MDFNNPCMD_INTEGRITY, NULL, 0);
                           EndNBTCPReceive(client);
                           StartNBTCPReceive(client, NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
                          }
			  else
			  {
                           EndNBTCPReceive(client);

			   if(len)
			    StartNBTCPReceive(client, NBTCP_COMMAND | cmd,len);
			   else	/* Woops.  Client probably tried to send a text message of 0 length. Or maybe a 0-length cheat file?  Better be safe! */
			    StartNBTCPReceive(client, NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			  }
			 }
			}
			return(1);

    case NBTCP_COMMAND: 
			{
			 uint32 len = client->nbtcplen;
			 uint32 tocmd = client->nbtcptype & 0xFF;

                         if(tocmd == MDFNNPCMD_INTEGRITY_RES)
                         {
                          printf("%s\n", md5_asciistr(client->nbtcp));
                         }
                         else if(tocmd == MDFNNPCMD_ECHO)
                         {
                          uint32 newlen = len;

                          if(newlen > 256)
                           newlen = 256;

                          SendCommand(client, tocmd, client->nbtcp, newlen);
                         }
			 else if(tocmd == MDFNNPCMD_SETNICK)
			 {
			  ParseAndSetNickname(client, client->nbtcp, len, true);
			 }
			 else if(tocmd == MDFNNPCMD_TEXT)
			 {
                          const uint32 nicklen = strlen(client->nickname);
                          uint8 mewmewbuf[4 + MaxNickLen + len];

                          en32(mewmewbuf, nicklen);

                          memcpy(mewmewbuf + 4, client->nickname, nicklen);
                          memcpy(mewmewbuf + 4 + nicklen, client->nbtcp, len);
                          SendToAllInSameGame(client, tocmd, mewmewbuf, 4 + nicklen + len);
			 }
			 else
			  SendToAllInSameGame(client, tocmd, client->nbtcp, len);

                         EndNBTCPReceive(client);
			 StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			}
			return(1);

    case NBTCP_LOGINLEN:
			{
			 uint32 len = de32(client->nbtcp);

			 // + 8192 because the client wouldn't know our maximum nickname length setting on the server beforehand, and to handle
			 // versions of Mednafen that send garbage(not technically garbage, but never really implemented on the server side, and a poor
			 // solution to the problem of input device selection anyway, so GARBAGE).
			 if(len > (sizeof(login_data_t) + MaxNickLen + 8192) || len < sizeof(login_data_t))
			 {
			  printf("Login len(%u) out of range.\n", len);
			  throw(1);
			 }

			 EndNBTCPReceive(client);
			 StartNBTCPReceive(client, NBTCP_LOGIN, len);
			}
			return(1);

    case NBTCP_LOGIN:
			{
			 const login_data_t *ld = (const login_data_t *)client->nbtcp;
			 const uint32 len = client->nbtcplen;
			 const uint32 pd_name_len = de32(ld->pd_name_len);
			 uint32 nickname_len = 0;
			 uint8 emu_id[64];

			 if(pd_name_len < (len - sizeof(login_data_t)))
			  nickname_len = len - (sizeof(login_data_t) + pd_name_len);

			 client->local_players = ld->local_players;
			 client->protocol_version = ld->protocol_version;
			 client->total_controllers = ld->total_controllers;
			 client->merge_target = ld->merge_target;

			 // Don't even send error messages for these, because the
			 // message sending code relies on the parameters that are invalid x_x
                         if(client->protocol_version < 3 && client->total_controllers > 8)
                         {
                          printf("That number of controllers(%u) isn't supported with that protocol version.\n", client->total_controllers);
                          throw(1);
                         }

                         if(client->total_controllers > MaxControllersPerGame)
                         {
                          printf("That number of controllers(%u) isn't supported.\n", client->total_controllers);
                          throw(1);
                         }

			 if(!client->protocol_version)
			 {
			  puts("Protocol 0 not supported.");
			  throw(1);
		         }

			 client->sendqalloced = ServerConfig.MinSendQSize;
			 client->sendqlen = 0;
			 client->sendq = (uint8 *)calloc(client->sendqalloced, 1);

			 if(!client->controller_data_size || !client->controller_data_offset || !client->sendq)
			 {
			  puts("Memory allocation failure!");
			  throw(1);
			 }

			 if(client->protocol_version < 2)
			 {
			  for(int x = 0; x < client->total_controllers; x++)
			  {
			   client->controller_data_size[x] = 1;
			   client->controller_data_offset[x] = x;
			  }
			  client->total_controllers_data_size = client->total_controllers;
			 }
		 	 else
			 {
			  int zeindex = 0;

			  for(int x = 0; x < client->total_controllers; x++)
			  {
			   client->controller_data_size[x] = ld->controller_data_size[x];
			   client->controller_data_offset[x] = zeindex;
			   zeindex += client->controller_data_size[x];
			  }
			  client->total_controllers_data_size = zeindex;
		 	 }

			 // Hack so we can always be able to encode our command length...
			 if(client->total_controllers_data_size < 4) 
			  client->total_controllers_data_size = 4;

			 if(client->total_controllers_data_size > MaxTotalControllersDataSize)
			 {
			  // TODO: Print information about client.
			  puts("Exceeded MaxTotalControllersDataSize");
			  throw(1);
			 }

			 client->local_controllers_data_size = 0;

                         if((client->merge_target - 1) >= client->total_controllers)
                         {
			  TextToClient(client, "Invalid merge target: player %d", client->merge_target);
			  throw(1);
			 }

                         if(ServerConfig.Password)
			 {
                          if(memcmp(ServerConfig.Password, ld->password, 16))
                          {
                           TextToClient(client,"Invalid server password.");
                           throw(1);
                          }
			 }

			 if(client->local_players < 1 || client->local_players > client->total_controllers)
			 {
			  puts("Invalid number of local players.");
			  TextToClient(client,"Invalid number(%d) of local players!",client->local_players);
			  throw(1);
			 }

			 memset(emu_id, 0, 64);

			 if(client->protocol_version == 3 && pd_name_len && pd_name_len <= 64 && ((int64_t)sizeof(login_data_t) + nickname_len + pd_name_len) <= len)
			 {
			  strncpy((char *)emu_id, (const char *)ld->nickname + nickname_len, pd_name_len);

			  for(int i = 0; i < 64; i++)
			  {
			   if(emu_id[i] > 0 && emu_id[i] < 0x20)
			    emu_id[i] = 0x20;
			  }
			 }

			 // Needs to be before nickname set, but we'll fix this eventually.
                         AddClientToGame(client, ld->gameid, emu_id);

                         ParseAndSetNickname(client, ld->nickname, nickname_len);

			 printf("Client(protocol %d) %d assigned to game %d as player %s <%s> - EmuID=%.64s\n", client->protocol_version, client->id, (int)(client->game - Games), MakeMPSString(client), client->nickname, emu_id);

			 GameEntry *tg = client->game;

			 // First, announce any existing clients to the new client.
                         for(int x = 0; x < MaxClientsPerGame && tg->TotalControllers; x++)
                         {
                          if(tg->Clients[x] && tg->IsUnique[x])
                          {
                           AnnouncePlayer(client, tg->Clients[x]);
                          }
                         }

			 // Now, announce this new client to other players.
			 // (Do this in a separate loop so we need less convoluted error handling)
                         for(int x = 0; x < MaxClientsPerGame && tg->TotalControllers; x++)
                         {
                          if(tg->Clients[x] && tg->IsUnique[x])
                          {
                           if(tg->Clients[x] != client)
                           {
                            try
                            {
                             AnnouncePlayer(tg->Clients[x], client);
                            }
                            catch(int i)
                            {
                             KillClient(tg->Clients[x]);
                            }
                           }
                          }
                         }

			// Old way, crashes if someone joins and leaves at the right times.
			// Could also result in a client disconnect notice being sent to the new client without a client connect notice
			// before it.
#if 0
			 for(int x = 0; x < MaxClientsPerGame && tg->TotalControllers; x++)
			 {
			  if(tg->Clients[x] && tg->IsUnique[x])
			  {
			   if(tg->Clients[x] != client)
			   {
			    try
			    {
			     AnnouncePlayer(tg->Clients[x], client);
			    }
			    catch(int i)
			    {
			     KillClient(tg->Clients[x]);
			    }
			    AnnouncePlayer(client, tg->Clients[x]);
			   }
			   else
			   {
			    AnnouncePlayer(client, client);
			   }
			  }
			 }
#endif
			 // Calculate the size of this client's local controllers data size.
			 {
			  uint32 localmask = MakeMPS(client, NULL);
			  client->local_controllers_data_size = 0;

			  for(int x = 0; x < client->total_controllers; x++)
			   if(localmask & (1 << x))
			    client->local_controllers_data_size += client->controller_data_size[x];
			 }
			} 
			EndNBTCPReceive(client);
			StartNBTCPReceive(client,NBTCP_UPDATEDATA,client->local_controllers_data_size + 1);
			return(1);
   }
  }
 }
 return(0);
}

int ListenSocket;

// Remove any control characters or reserved characters from the nickname.
static void CleanNick(char *nick)
{
 for(unsigned int x = 0; x < strlen(nick); x++)
  if(nick[x] == '<' || nick[x] == '>' || nick[x] == '*' || (nick[x] < 0x20))
   nick[x] = 0;
}

static bool NickUnique(ClientEntry *client, const char *newnick)
{
 int x;
 GameEntry *game = client-> game;

 for(x = 0; x < MaxClientsPerGame; x++)
  if(game->Clients[x] && client != game->Clients[x])
   if(!strcasecmp(newnick, game->Clients[x]->nickname))
    return(0);
 return(1);
}

static void ParseAndSetNickname(ClientEntry *client, const uint8 *data, uint32 data_len, bool is_change)        // is_change is false on initial login, true if trying to change nickname after login.
{
 char newnick[MaxNickLen + 1];

 if(data_len > (uint32)MaxNickLen)
  data_len = MaxNickLen;

 newnick[data_len] = 0;
 memcpy(newnick, data, data_len);

 CleanNick(newnick);

 if(!is_change) // Login nickname set
 {
  if(!strlen(newnick) || !NickUnique(client, newnick))
   strcpy(newnick, MakeMPSString(client));

  strcpy(client->nickname, newnick);
 }
 else
 {
  if(!strlen(newnick))
   strcpy(newnick, MakeMPSString(client));

  if(NickUnique(client, newnick))
  {
   uint8 ncdata[MaxNickLen + 1 + MaxNickLen + 1];
   uint32 ncdata_len;

   if(!strlen(newnick))
    strcpy(newnick, MakeMPSString(client));

   ncdata_len = strlen(client->nickname) + 1 + strlen(newnick) + 1;

   memcpy(ncdata, client->nickname, strlen(client->nickname));
   ncdata[strlen(client->nickname)] = '\n';
   strcpy((char*)ncdata + strlen(client->nickname) + 1, newnick);

   strcpy(client->nickname, newnick);

   SendToAllInSameGame(client, MDFNNPCMD_NICKCHANGED, ncdata, ncdata_len);
  }
 }
}

static int MakeSendTCP(ClientEntry *client, const uint8 *data, uint32 len) throw(int)
{
 // If there's no data in our software sendq, try sending this data without using the soft sendq.
 if(!client->sendqlen)
 {
  int32 sent = send(client->TCPSocket, data, len, MSG_NOSIGNAL);

  if(sent < 0)
  {
   if(errno == EAGAIN || errno == EWOULDBLOCK)
    sent = 0;
   else
   {
    //puts("Fatal error on send()");
    throw(1);
   }
  }

  assert((uint32)sent <= len);

  // ...and adjust the data pointer and length appropriately.
  data += sent;
  len -= sent;

  // If we sent all the data, no need to go through the software sendq code at all~
  if(!len)
   return(1);
 }




 assert(client->sendq);


 if((client->sendqlen + len) > client->sendqalloced)
 {
  uint32 new_alloc_size;

  new_alloc_size = client->sendqlen + len;

  if(new_alloc_size > ServerConfig.MaxSendQSize)
  {
   printf("Dropped client %d on game %d due to exceeding MaxSendQSize by %u bytes\n", client->id, (int)(client->game - Games), ServerConfig.MaxSendQSize - new_alloc_size);
   throw(1);
  }

  new_alloc_size += new_alloc_size >> 1;
  if(new_alloc_size > ServerConfig.MaxSendQSize)
   new_alloc_size = ServerConfig.MaxSendQSize;

  if(!(client->sendq = (uint8 *)realloc(client->sendq, new_alloc_size)))
  {
   printf("Dropped client due to realloc failure: %m\n");
   throw(1);
  }
  client->sendqalloced = new_alloc_size;
 }

 //puts("SendQ path");

 // Append new data to the end of the sendq
 memcpy(client->sendq + client->sendqlen, data, len);
 client->sendqlen += len;

 // Now try to send all the data in the sendq
 int32 sent = send(client->TCPSocket, client->sendq, client->sendqlen, MSG_NOSIGNAL);

 if(sent < 0) 
 {
  if(errno == EAGAIN || errno == EWOULDBLOCK)
   sent = 0;
  else
  {
   //puts("Fatal error on send()");
   throw(1);
  }
 }

 if(sent)
 {
  uint32 new_alloc_size;

  client->sendqlen -= sent;
  memmove(client->sendq, client->sendq + sent, client->sendqlen);


  new_alloc_size = client->sendqlen;
  if(new_alloc_size < ServerConfig.MinSendQSize)
   new_alloc_size = ServerConfig.MinSendQSize;

  if(!(client->sendq = (uint8 *)realloc(client->sendq, new_alloc_size)))
  {
   printf("Dropped client due to realloc failure: %m\n");
   throw(1);
  }
  client->sendqalloced = new_alloc_size;
 }

 return(1);
}

static void SendCommand(ClientEntry *client, uint8 cmd, const uint8 *data, uint32 len) throw(int)
{
 uint8 poo[client->total_controllers_data_size + 1];

 memset(poo, 0, client->total_controllers_data_size);

 poo[client->total_controllers_data_size] = cmd;

 if((cmd & 0x80) && data != NULL)
  en32(poo, len);

 MakeSendTCP(client, poo, client->total_controllers_data_size + 1);

 if((cmd & 0x80) && data != NULL)
  MakeSendTCP(client, data, len);
}

static void SendToAllInSameGame(ClientEntry *client, int cmd, const uint8 *data, uint32 len) throw(int)
{
 GameEntry *game = client->game;	// Cache the game pointer, since the client data structure can be destroyed.
 bool this_client_destroyed = false;
 uint64 pending_destruction = 0;

 for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
 {
  if(!game->Clients[x] || !game->IsUnique[x]) continue;

  try
  {
   SendCommand(game->Clients[x], cmd, data, len);
  }
  catch(int i)
  {
   pending_destruction |= 1ULL << x;
  }
 }

 // Do it this way so we won't send disconnection notices and text messages regarding the same client out-of-order.
 for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
 {
  if(!game->Clients[x] || !game->IsUnique[x]) continue;

  if(pending_destruction & (1ULL << x))
  {
   if(game->Clients[x] == client)
    this_client_destroyed = true;

   KillClient(game->Clients[x]);
  }
 }


 if(this_client_destroyed)
  throw(1);
}

static void TextToClient(ClientEntry *client, const char *fmt, ...) throw(int)
{
 uint8 moo[4 + 1024]; // nick len(=0) and message
 va_list ap;

 va_start(ap,fmt);
 vsnprintf((char*)moo + 4, 1024, fmt, ap);
 va_end(ap);
 
 en32(moo, 0); // nick len of 0 to signify this is a server message
 SendCommand(client, 0x90, moo, 4 + strlen((char*)moo + 4));
}

static void AnnouncePlayer(ClientEntry *client, ClientEntry *newclient) throw(int)
{
 uint32 mps, mergedto;
 uint8 data[4 + 4 + MaxNickLen];
 uint32 datalen;

 mps = MakeMPS(newclient, &mergedto);

 if(client->protocol_version >= 3)
 {
  datalen = 4 + 4 + strlen(newclient->nickname);

  en32(&data[0], mps);
  en32(&data[4], mergedto);

  memcpy(data + 8, newclient->nickname, strlen(newclient->nickname));
 }
 else
 {
  datalen = 1 + 1 + strlen(newclient->nickname);

  data[0] = mps;
  data[1] = mergedto;

  memcpy(data + 2, newclient->nickname, strlen(newclient->nickname));
 }

 if(client == newclient)
  SendCommand(client, MDFNNPCMD_YOUJOINED, data, datalen);
 else
  SendCommand(client, MDFNNPCMD_PLAYERJOINED, data, datalen);
}

static void AnnouncePlayerLeft(ClientEntry *client, ClientEntry *newclient, uint32 mps, uint32 mergedto) throw(int)
{
 uint8 data[4 + 4 + MaxNickLen];
 uint32 datalen;

 if(client->protocol_version >= 3)
 {
  datalen = 4 + 4 + strlen(newclient->nickname);

  en32(&data[0], mps);
  en32(&data[4], mergedto);

  memcpy(data + 8, newclient->nickname, strlen(newclient->nickname));
 }
 else
 {
  datalen = 1 + 1 + strlen(newclient->nickname);

  data[0] = mps;
  data[1] = mergedto;

  memcpy(data + 2, newclient->nickname, strlen(newclient->nickname));
 }

 if(client == newclient)
  SendCommand(client, MDFNNPCMD_YOULEFT, data, datalen);
 else
  SendCommand(client, MDFNNPCMD_PLAYERLEFT, data, datalen);
}

static void KillClient(ClientEntry *client)
{
 GameEntry *game;

 if(!client)
 {
  puts("Erhm, attempt to kill NULL client o_O");
  return;
 }

 if(client->TCPSocket == -1)
 {
  puts("Attempt to kill client multiple times.  The grim reaper has gone mad!");
  return;
 }

 game = client->game;
 if(game)
 {
  //
  // Calculate what the player controlled before we destroy that information.
  //
  uint32 mps, mergedto;

  mps = MakeMPS(client, &mergedto);

  // Remove the client from the game's player list BEFORE we do an announcement that the client has left.  Which should be obvious, but we used to do it backwards
  // and wonkiness randomly ensued. :/
  //
  // And uncomment out this code immediately below if we ever have a "nice" quit handshake rather than a hard disconnect.
  // try
  // {
  //  AnnouncePlayerLeft(client, client, mps, mergedto);
  // }
  // catch(int i)
  // {
  //  // Don't do anything here, since we're already on the way to kill the client.  KILL IT DEAD.
  // }

  for(int x = 0; x < MaxClientsPerGame; x++)
   if(game->Clients[x])
    if(game->Clients[x] == client)
     game->Clients[x] = NULL;

  // Now announce to the remaining players that the client has left.
  for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
  {
   if(!game->Clients[x] || !game->IsUnique[x])
    continue;

   try
   {
    AnnouncePlayerLeft(game->Clients[x], client, mps, mergedto);
   }
   catch(int i)
   {
    KillClient(game->Clients[x]);
   }
  }

  time_t curtime = time(0);
  printf("Player <%s> disconnected from game %d on %s", client->nickname, (int)(game - Games), ctime(&curtime));


  if(game->TotalControllers)	// Make sure the game still exists to destroy, since we could have killed it in a recursive call to KillClient() above.
  {
   bool AnyClientsLeft = false;		// Any clients left connected to this game?

   for(int x = 0; x < MaxClientsPerGame && !AnyClientsLeft; x++)
    if(game->Clients[x])
     AnyClientsLeft = true;

   if(!AnyClientsLeft)
   {
    printf("Game %d destroyed.\n", (int)(game-Games));
    memset(game, 0, sizeof(GameEntry));
    game = 0;
   }
  }
 }
 else
 {
  time_t curtime = time(0);
  printf("Unassigned client %d disconnected on %s",client->id, ctime(&curtime));
 }

 if(client->sendq)
  free(client->sendq);

 if(client->nbtcp)
  free(client->nbtcp);
 
 if(client->TCPSocket != -1)
  close(client->TCPSocket);
 memset(client, 0, sizeof(ClientEntry));
 client->TCPSocket = -1;
}

static void AddClientToGame(ClientEntry *client, const uint8 id[16], const uint8 *emu_id) throw(int)
{
 GameEntry *game,*fegame;

 retry:

 game = NULL;
 fegame = NULL;

 /* First, find an available game. */
 for(int wg = 0; wg < ServerConfig.MaxClients; wg++)
 {
  if(!Games[wg].TotalControllers && !fegame)
  {
   if(!fegame)
    fegame=&Games[wg];
  }
  else if(Games[wg].TotalControllers && !memcmp(Games[wg].id,id,16)) /* A match was found! */
  {
   game = &Games[wg];
   break;
  }
 }

 if(game)
 {
  if(game->ProtocolVersion != client->protocol_version)
  {
   TextToClient(client, "Protocol version mismatch.  Game: %u, You: %u", game->ProtocolVersion, client->protocol_version);
   throw(1);
  }

  if(memcmp(game->EmulatorID, emu_id, 64))
  {
   TextToClient(client, "Emulator (version) mismatch.  Game: %.64s, You: %.64s", game->EmulatorID, emu_id);
   throw(1);
  }

  if(game->TotalControllers != client->total_controllers || game->TotalControllersDataSize != client->total_controllers_data_size)
  {
   TextToClient(client, "Hunh?  Controllers mismatch...");
   throw(1);
  }
 }
 if(!game) /* Hmm, no game found.  Guess we'll have to create one. */
 {
  game=fegame;
  printf("Game %d added\n",(int)(game-Games));
  memset(game, 0, sizeof(GameEntry));

  game->TotalControllers = client->total_controllers; // total_controllers is validated in the user login code
  game->TotalControllersDataSize = client->total_controllers_data_size;

  memcpy(game->ControllerDataSize, client->controller_data_size, game->TotalControllers * sizeof(int));
  memcpy(game->ControllerDataOffset, client->controller_data_offset, game->TotalControllers * sizeof(int));

  game->fps = 1008307711; // NES NTSC FPS magical number from beyond the moon, the default.
  game->last_time = MBL_Time64();
  game->ProtocolVersion = client->protocol_version;
  memcpy(game->EmulatorID, emu_id, 64);
  memcpy(game->id, id, 16);
 }


 for(int n = 0; n < MaxClientsPerGame; n++)
  if(game->Clients[n])
  {
   try
   {
    SendCommand(game->Clients[n], 0x81, NULL, 0);
    break;
   }
   catch(int i)
   {
    KillClient(game->Clients[n]);
    /* All right, now the client we sent the request to is killed.  I HOPE YOU'RE HAPPY! */
    /* Since killing a client can destroy a game, we'll need to see if the game was trashed.  
       If it was, then "goto retry", and try again.  Ugly, yes.  I LIKE UGLY.
    */
    if(!game->TotalControllers)
     goto retry;
   }
  }

 int instancecount = client->local_players;

 for(int n = 0; n < MaxClientsPerGame && instancecount; n++)
 {
  // If we're in lurker territory, and we already assigned this client an instance in the game, don't assign any more.
  if((n >= game->TotalControllers) && (instancecount < client->local_players))
  {
   // Set instancecount to 0 so we don't trigger the "game is full" error below this loop.
   instancecount = 0;
   break;
  }

  if(!game->Clients[n])
  {
   game->Clients[n] = client;

   if(instancecount == client->local_players)
    game->IsUnique[n] = 1;
   else
    game->IsUnique[n] = 0;

   instancecount--;
  }
 }

 /* Game is full. */
 if(instancecount)
 {
  // Remove any instances of this client added to the game.
  for(int n = 0; n < MaxClientsPerGame; n++)
   if(game->Clients[n] == client)
    game->Clients[n] = NULL;

  TextToClient(client, "Sorry, game is full.  %d instance(s) tried, %d available.", client->local_players, client->local_players - instancecount);
  throw(1);
 }

 client->game = game;
}

int main(int argc, char *argv[])
{
 union
 {
  #ifdef HAVE_IPV6
  struct sockaddr_in6 sockin6;
  #endif
  struct sockaddr_in sockin;

  struct sockaddr saddr;
 };
 socklen_t saddr_len;

 #ifdef HAVE_IPV6
 bool UsingIPV6 = true;
 #endif

 if(argc < 2)
 {
  if(strchr(argv[0],' '))
   printf("Usage: \"%s\" <configfile>\n",argv[0]);
  else
   printf("Usage: %s <configfile>\n",argv[0]);
  exit(-1);
 }

 printf("Starting Mednafen-Server %s\n\n", VERSION);

 if(!LoadConfig(argv[1]))
 {
  puts("Error loading configuration file");
  exit(-1);
 }

 Games = (GameEntry *)calloc(ServerConfig.MaxClients, sizeof(GameEntry));
 AllClients = (ClientEntry *)calloc(ServerConfig.MaxClients, sizeof(ClientEntry));

 for(int x = 0; x < ServerConfig.MaxClients; x++)
 {
  AllClients[x].TCPSocket = -1;
 }

 /* First, we need to create a socket to listen on. */
#ifdef HAVE_IPV6
 if(UsingIPV6)
 {
  if(-1 == (ListenSocket = socket(AF_INET6, SOCK_STREAM, 0)))
  {
   printf("IPv6 socket creation failed(%m), trying IPv4 instead.\n");
   UsingIPV6 = false;
  }
 }
#endif

#ifdef HAVE_IPV6
 if(!UsingIPV6)
#endif
 {
  if(-1 == (ListenSocket = socket(AF_INET, SOCK_STREAM, 0))) 
  {
   printf("Error creating socket: %m\n");
   exit(-1);
  }
 }

 /* Set send buffer size to 262,144 bytes. */
 int sndbufsize = 262144;
 if(setsockopt(ListenSocket, SOL_SOCKET, SO_SNDBUF, &sndbufsize, sizeof(int)))
  puts("Send buffer size set failed.");

 int tcpopt = 1;
 if(setsockopt(ListenSocket, SOL_TCP, TCP_NODELAY, &tcpopt, sizeof(int)))
 {
  puts("Nodelay fail");
  exit(-1);
 }

#ifdef HAVE_IPV6
 if(UsingIPV6)
 {
  saddr_len = sizeof(sockin6);

  memset(&sockin6, 0, sizeof(sockin6));
  memcpy(&sockin6.sin6_addr, &in6addr_any, sizeof(in6_addr));	// = IN6ADDR_ANY_INIT;
  sockin6.sin6_family = AF_INET6;
  sockin6.sin6_port = htons(ServerConfig.Port);
 }
 else
#endif
 {
  saddr_len = sizeof(sockin);

  memset(&sockin, 0, sizeof(sockin));
  sockin.sin_addr.s_addr = INADDR_ANY;
  sockin.sin_family = AF_INET;
  sockin.sin_port = htons(ServerConfig.Port);
 }

 if(bind(ListenSocket, &saddr, saddr_len))
 {
  printf("%s\n", strerror(errno));
  puts("Bind error.");
  exit(-1);
 }

 if(listen(ListenSocket, 8))
 {
  puts("Listen error.");
  exit(-1);
 }
 printf("Listening on port %d.\n",ServerConfig.Port);

 /* We don't want to block on accept() */
 fcntl(ListenSocket, F_SETFL, fcntl(ListenSocket, F_GETFL) | O_NONBLOCK);

 #if defined(HAVE_MLOCKALL) && defined(MCL_CURRENT)
 if(mlockall(MCL_CURRENT) != 0)
  printf("Note: mlockall(MCL_CURRENT) failed: %m (not a major problem; see the documentation if curious)\n");
 #endif

 /* Now for the BIG LOOP. */
 while(1)
 {
  int n;

  for(n=0; n<ServerConfig.MaxClients; n++)
  {
   if(AllClients[n].TCPSocket != -1) 
    continue;
  
   socklen_t saddr_accept_len = saddr_len;

   if((AllClients[n].TCPSocket = accept(ListenSocket, &saddr, &saddr_accept_len)) != -1)
   {
    /* We have a new client.  Yippie. */

    fcntl(AllClients[n].TCPSocket, F_SETFL, fcntl(AllClients[n].TCPSocket, F_GETFL) | O_NONBLOCK);

    AllClients[n].timeconnect = time(0);
    AllClients[n].id = n;

#ifdef HAVE_IPV6
    if(UsingIPV6)
    {
     char tbuf[256];
     printf("Client %d connecting from %s on %s", n, inet_ntop(AF_INET6, &sockin6.sin6_addr, tbuf, sizeof(tbuf)), ctime(&AllClients[n].timeconnect));
    }
#endif
    else
    {
     printf("Client %d connecting from %s on %s", n, inet_ntoa(sockin.sin_addr), ctime(&AllClients[n].timeconnect));
    }

    StartNBTCPReceive(&AllClients[n], NBTCP_LOGINLEN, 4);
   }
  }  

  /* Check for users still in the login process(not yet assigned a game). BOING */
  time_t curtime = time(0);
  for(n = 0; n < ServerConfig.MaxClients; n++)
   try
   {
    if(AllClients[n].TCPSocket != -1 && !AllClients[n].game)
    {
     if((AllClients[n].timeconnect + ServerConfig.ConnectTimeout) < curtime)
     {
      KillClient(&AllClients[n]);
     }
     else
      while(CheckNBTCPReceive(&AllClients[n])) {};
    }
   }
   catch(int i) 
   {
    KillClient(&AllClients[n]);
   }
  int whichgame;
  for(whichgame = 0; whichgame < ServerConfig.MaxClients; whichgame ++)
  {
   if(!Games[whichgame].fps)
    continue;

   int64 last_time_inc = (int64)65536 * 256 * 1000000 / Games[whichgame].fps;
   const int64 CurGameTimeUS = MBL_Time64();

   if((Games[whichgame].last_time + last_time_inc) > CurGameTimeUS)
    continue;

   Games[whichgame].last_time += last_time_inc;
   //printf("%lld\n", last_time_inc);

   /* Now, the loop to get data from each client.  Meep. */
   for(n = 0; n < MaxClientsPerGame && Games[whichgame].TotalControllers; n++)
   {
    try
    {
     ClientEntry *client = Games[whichgame].Clients[n];
     if(!client || !Games[whichgame].IsUnique[n]) 
      continue;

     while(CheckNBTCPReceive(client)) {};

     if((client->last_receive_time + (int64)ServerConfig.IdleTimeout * 1000 * 1000) < CurGameTimeUS)
     {
      printf("Client %d idle timeout disconnect.\n", client->id);
      throw(1);
     }
    } // catch
    catch(int i)
    {
     KillClient(Games[whichgame].Clients[n]);
    }
   } // A games clients

   if(!Games[whichgame].fps) // Oops, client exiting killed it.
   {
    continue;
   }

   uint8 merge_data_buf[Games[whichgame].TotalControllersDataSize + 1];

   memcpy(merge_data_buf, Games[whichgame].joybuf, sizeof(merge_data_buf));

   /* Now merge!!! */
   for(n = 0; n < Games[whichgame].TotalControllers; n++)
   {
    if(!Games[whichgame].Clients[n] || !Games[whichgame].IsUnique[n]) 
     continue;

    if(Games[whichgame].Clients[n]->merge_target)
    {
     int offshoo = Games[whichgame].ControllerDataOffset[n];
     int offshoomt = Games[whichgame].ControllerDataOffset[ Games[whichgame].Clients[n]->merge_target - 1 ];

     for(int zoom = 0; zoom < Games[whichgame].ControllerDataSize[n]; zoom++)
     {
      merge_data_buf[offshoomt + zoom] |= Games[whichgame].joybuf[offshoo + zoom];
     }
    }
   }

   /* Now we send the data to all the clients. */
   for(n = 0; n < MaxClientsPerGame && Games[whichgame].TotalControllers; n++)
   {
    if(!Games[whichgame].Clients[n] || !Games[whichgame].IsUnique[n])
     continue;

    try
    {
     MakeSendTCP(Games[whichgame].Clients[n], merge_data_buf, sizeof(merge_data_buf));
    }
    catch(int i)
    {
     KillClient(Games[whichgame].Clients[n]);
    }
   } // A game's clients
  } // Games

  // Now, determine the amount of time we should sleep, and sleep it.
  // TODO: Consider using select() with a timeout in the future, so that throughput will be higher
  //       when running on very fast links with small local OS TCP receive and send buffers.
  int64 min_sleep = 25000;
  int64 curgametime = MBL_Time64(); // Don't use cached game time from above, otherwise we could oversleep.
  for(whichgame = 0; whichgame < ServerConfig.MaxClients; whichgame ++)
  {
   if(!Games[whichgame].fps)
    continue;
   int64 last_time_inc = (int64)65536 * 256 * 1000000 / Games[whichgame].fps;
   int64 sandwich = (Games[whichgame].last_time + last_time_inc) - curgametime;
   if(sandwich <= 0)
   {
    min_sleep = 0;
    break;
   }
   else if(sandwich < min_sleep) min_sleep = sandwich;
  }
  if(min_sleep > 0)
   MBL_Sleep64(min_sleep);
  //printf("%lld, %lld\n", min_sleep, MBL_Time64() - curgametime);
 } // while(1)
}
